/* codes for HTTP request methods */
#define RM_NONE -1
#define RM_GET  0
#define RM_HEAD 1
#define RM_POST 2

/* Our http request structure */
struct rq {
    char *host;		/* Host name or IP to connect to.  Never NULL. */ 
    int port;		/* Port number to connect to */
    char *path;		/* Path of resource (always starts with '/') */
    char *query;	/* Query string 'name=value&name=value', encoded */
    char method;	/* Method to use, eg RM_POST */
    char *user;		/* User name (if authenticating, NULL otherwise) */
    char *password;	/* Password (if authenticating, NULL otherwise) */
    int redirect;	/* True if we should allow redirects */
};

/* Our http response structure */
struct rp {
    int status;		/* Response status code */
    char *reason;	/* Reason string */
    char *redirect;	/* If not NULL, Location redirect URI */
    char *authrealm;	/* If not NULL, Basic Authentication realm */
    char *type;		/* If not NULL, Content Type */
    char *content;	/* If not NULL, Decoded Content */
};

/* Describe a proxy server */
struct px {
    char *host;		/* Proxy Hostname */
    int port;		/* Proxy port number */
};

void free_request(struct rq *req);
void free_response(struct rp *resp);
void free_proxy(struct px *proxy);
struct rq *uri_request(char *uri, int method);
int tcp_open(char *hostname, int portnum);
struct rp *execute_request(struct rq *req, struct px *proxy);
struct px *parse_proxy(char *str);
int method_code(char *str);
